/*
 * This file is part of the flashrom project.
 *
 * Copyright (C) 2023 Celliwig
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/*
 * Contains the routines for the Nuvaton/Winbond WPCE775, supports:
 * 	Acer Aspire 8730g
 */

#define PCI_VENDOR_ID_INTEL		0x8086
#define PCI_HW_ID_INTEL_LPC		0x0601

#define WPCE775_SUPERIO_PORT1		0x2e
#define WPCE775_SUPERIO_PORT2		0x4e
#define WPCE775_SUPERIO_PORT3		0x66

#define WPCE775_KBC_DATA_PORT		0x60
#define WPCE775_KBC_STATUS_PORT		0x64
#define WPCE775_KBC_COMMAND_PORT	0x64
#define WPCE775_UNDEFINED_PORT		0x00ED

#define WPCE775_MMIO_ADDR		0xff800000
#define	WPCE775_MMIO_LEN		0x100
#define WPCE775_LOCK_ADDR		0xffa00000
#define	WPCE775_LOCK_LEN		0x200000
#define WPCE775_ROM_ADDR		0xffe00000
#define	WPCE775_ROM_LEN			0x200000

// Init data copied from ROM image
static const uint8_t wpce775_inittable[0xe][0xe] = {
				{ 0x06, 0x06, 0x3B, 0x16, 0x02, 0x03, 0x20, 0x05, 0x90, 0xC7, 0x01, 0x08, 0xEF, 0x13 },
				{ 0x06, 0x50, 0x3B, 0x16, 0x02, 0x03, 0x20, 0x05, 0x90, 0x60, 0x01, 0x01, 0xBF, 0x8E },
				{ 0x06, 0x50, 0x3B, 0x16, 0x02, 0x03, 0x20, 0x05, 0x90, 0x60, 0x01, 0x01, 0xBF, 0x80 },
				{ 0x06, 0x06, 0x3B, 0x16, 0x02, 0x03, 0x20, 0x05, 0x90, 0xC7, 0x01, 0x08, 0xC2, 0x13 },
				{ 0x06, 0x06, 0x3B, 0x16, 0x02, 0x03, 0x20, 0x05, 0x9F, 0xC7, 0x01, 0x08, 0x1F, 0x45 },
				{ 0x06, 0x06, 0x3B, 0x16, 0x02, 0x03, 0x20, 0x05, 0x90, 0xC7, 0x01, 0x08, 0xEF, 0x14 },
				{ 0x06, 0x06, 0x3B, 0x16, 0x02, 0x03, 0x20, 0x05, 0x9F, 0xC7, 0x01, 0x08, 0x20, 0x15 },
				{ 0x06, 0x06, 0x3B, 0x16, 0x02, 0x03, 0x20, 0x05, 0x9F, 0xC7, 0x01, 0x08, 0x1F, 0x46 },
				{ 0x06, 0x06, 0x3B, 0x16, 0x02, 0x03, 0x20, 0x05, 0x90, 0xC7, 0x01, 0x08, 0xC2, 0x14 },
				{ 0x06, 0x50, 0x3B, 0x16, 0x02, 0x03, 0x20, 0x05, 0x90, 0x60, 0x01, 0x01, 0xBF, 0x41 },
				{ 0x06, 0x06, 0x3B, 0x16, 0x02, 0x03, 0x20, 0x05, 0x90, 0xC7, 0x01, 0x08, 0x1C, 0x14 },
				{ 0x06, 0x06, 0x3B, 0x16, 0x02, 0x03, 0x20, 0x05, 0x90, 0xC7, 0x01, 0x08, 0xEF, 0x15 },
				{ 0x06, 0x06, 0x3B, 0x14, 0x02, 0x03, 0x20, 0x05, 0x9F, 0xC7, 0x01, 0x08, 0x1F, 0x47 },
				{ 0x06, 0x06, 0x3B, 0x16, 0x02, 0x03, 0x20, 0x05, 0x90, 0xC7, 0x01, 0x08, 0xC2, 0x15 },
				};

struct wpce775_superio {
	uint16_t apci_io_base;		// ACPI I/O port base address
	uint8_t acpi_smi_prev;		// ACPI previous value for SMI enable
	uint16_t mcu1_io_base;		// MCU 1 I/O port base address
	void *mcu2_mmio_dev;		// MMIO for MCU 2
	uint32_t mcu2_mmio_base;	// MMIO base address for MCU 2
	uint32_t mcu2_mmio_length;	// MMIO address length for MCU 2
	uint16_t mcu2_io_base;		// MCU 2 I/O port base address
	void *lock_mmio_dev;		// MMIO for lock
	uint32_t lock_mmio_base;	// MMIO base address for lock
	uint32_t lock_mmio_length;	// MMIO address length for lock
	void *rom_mmio_dev;		// MMIO for ROM
	uint32_t rom_mmio_base;		// MMIO base address for ROM
	uint32_t rom_mmio_length;	// MMIO address length for ROM
	struct pci_dev *pci_lpc;	// LPC ISA bridge PCI device
};

uint8_t wpce775_sio_read(uint16_t port, uint8_t reg);
void wpce775_sio_write(uint16_t port, uint8_t reg, uint8_t data);
void wpce775_sio_mask(uint16_t port, uint8_t reg, uint8_t data, uint8_t mask);

uint8_t wpce775_kbc_disable(struct wpce775_superio *wpce775);
void wpce775_kbc_enable(struct wpce775_superio *wpce775);

bool wpce775_acpi_pwrbtnen(struct wpce775_superio *wpce775, bool btn_enable);
void wpce775_acpi_smistate_clear(struct wpce775_superio *wpce775);
void wpce775_acpi_smistate(struct wpce775_superio *wpce775, bool smi_restore);

bool wpce775_lpc_probe(struct wpce775_superio *wpce775);
bool wpce775_lpc_conf_chk(struct wpce775_superio *wpce775);
void wpce775_lpc_bios_write_enable(struct wpce775_superio *wpce775, bool write_enable);

bool wpce775_mcu1_io_init(struct wpce775_superio *wpce775);

bool wpce775_mcu2_mmio_init(struct wpce775_superio *wpce775);
bool wpce775_mcu2_mmio_open(struct wpce775_superio *wpce775);
void wpce775_mcu2_mmio_close(struct wpce775_superio *wpce775);

void wpce775_mcu2_mmio_status_check_loop(struct wpce775_superio *wpce775, uint8_t chk_mask, uint8_t chk_val, uint8_t dbg_code);
void wpce775_mcu2_mmio_status_runclear(struct wpce775_superio *wpce775, uint8_t dbg_code);
void wpce775_mcu2_mmio_mode_enter(struct wpce775_superio *wpce775, uint8_t arg1, uint8_t arg2, uint8_t arg3, uint8_t arg4, uint8_t arg5, uint8_t arg6, uint8_t arg7, uint8_t arg8);
void wpce775_mcu2_mmio_mode_exit(struct wpce775_superio *wpce775);
uint32_t wpce775_mcu2_mmio_mode_jedecid(struct wpce775_superio *wpce775);
bool wpce775_mcu2_mmio_mode_init(struct wpce775_superio *wpce775);
void wpce775_mcu2_mmio_mode_flash(struct wpce775_superio *wpce775);
void wpce775_mcu2_mmio_mode_erase(struct wpce775_superio *wpce775, uint32_t sector_addr);
void wpce775_mcu2_mmio_mode_program(struct wpce775_superio *wpce775, const uint8_t *buf, uint32_t rom_addr, uint8_t len);
void wpce775_mcu2_mmio_lock(struct wpce775_superio *wpce775, bool memory_protect);

bool wpce775_lock_mmio_open(struct wpce775_superio *wpce775);
void wpce775_lock_mmio_close(struct wpce775_superio *wpce775);

bool wpce775_rom_mmio_open(struct wpce775_superio *wpce775);
void wpce775_rom_mmio_close(struct wpce775_superio *wpce775);

void wpce775_access_lock(struct wpce775_superio *wpce775);
int wpce775_access_unlock(struct wpce775_superio *wpce775);

int wpce775_dumpICH9Reg(struct pci_dev *lpc);

static int wpce775_probe(struct flashctx *flash);
static int wpce775_read(struct flashctx *flash, uint8_t *buf, unsigned int addr, unsigned int len);
static int wpce775_write(struct flashctx *flash, const uint8_t *buf, unsigned int addr, unsigned int len);
static int wpce775_erase(struct flashctx *flash, unsigned int addr, unsigned int len);
static int wpce775_shutdown(void *data);
